.module/ram/abs=0      C1SPN;
.pagelength  59;                    { Works for .LST files with 550C }
.include    <system.k>;

{******************************************************************************
 *
 *      EZ-KIT Lite DSP Program for SOSSD
 *	c1spn.dsp  -  Generate a single 1000 Hz sine wave plus Gaussian
 *	white noise.
 *
 *	24 March 1999    RSL
 *
 *****************************************************************************}

{******************************************************************************
 *
 * GENERAL NOTE - The initialization routines, in this program, have greatly
 * reduced comments, relative to those in the shell dsp program (c1shell.dsp).
 * This make the flow of this program easier to follow, but removes the
 * detailed explanations from the code.  See the shell for more details.
 *
 ****************************************************************************}

{************************   CONSTANTS   *************************************}
.const		SIN_LEV=3277;		{ 0.1 in 1.15}
.const		SIN_DPH=1365;		{ Phase for 1000.24 Hz }
.const		RN_SEED=0x1234;		{ This can be anything }

{***********************   PROGRAM MEMORY     *******************************}
.var/pm         sin_coeff[5];
.init           sin_coeff: H#324000, H#005300, H#AACC00, H#08B700, H#1CCE00;

{********************   DATA MEMORY     *******************************}
.var/dm/circ	tx_buf[3];		{ SPORT0: Cmd + L data + R data }
.var/dm/circ 	rx_buf[3];		{ SPORT0: Status + L data + R data }
.var/dm/circ    init_cmds[13];		{ CODEC Initialization }
.var/dm		stat_flag;		{ CODEC Init status }

.var/dm		phase;			{ Cumulated phase of sine wave }
.var/dm		signal_save;

.var/dm		seed_msw;		{ The seed for random number }
.var/dm		seed_lsw;		{   generation, saved as 2x16 bits }

{********************  Interrupt vector table  *******************************}
        jump start;  rti; rti; rti;     {00: reset }
        rti;         rti; rti; rti;     {04: IRQ2 }
        rti;         rti; rti; rti;     {08: IRQL1 }
        rti;         rti; rti; rti;     {0c: IRQL0 }
	ar = dm(stat_flag);		{10: SPORT0 tx }
        ar = pass ar;			{  see if needs initialization }
        if eq rti;
        jump next_cmd;
	jump input_samples;             {14: SPORT0 rx }
                     rti; rti; rti;
        rti;         rti; rti; rti;     {18: IRQE }
        rti;         rti; rti; rti;     {1c: BDMA }
	rti;         rti; rti; rti;     {20: SPORT1 tx or IRQ1 }
        rti;         rti; rti; rti;     {24: SPORT1 rx or IRQ0 }
	rti;         rti; rti; rti;     {28: timer }
        rti;         rti; rti; rti;     {2c: power down }

{*****************************  START  ***************************************
 * Initialize process and drop into main loop.
 ****************************************************************************}
start:	imask=0;			{ Turn off all interrupts }
	call init0;			{ Instructions that simulate easily }
	call init1;			{ And those that do not }

{****************     Main Program Loop    ******************************
 *  Main processing loop.  Does all non-periodic (background) processing.
 *  This loop is interrupted at a 48KHz rate to send and receive Codec data.
 *************************************************************************}
again:	jump again;			{ Go round and round forever }

{***********************************************************************
 *                 SPORT0 Interrupt Handler                            *
 *  SPORT0 inputs from the CODEC put us here 48000 times per second.   *
 *  Routines here are those that need this periodic service.           *
 ***********************************************************************}
input_samples:
        ena sec_reg;                    { use secondary register bank }
	{ For details on the generation of the sine wave, see c1sin.dsp }
	ax1 = SIN_DPH;			{ Phase increment for oscillator }
	ay1 = dm(phase);		{ Last phase }
	ar = ax1 + ay1;			{ New phase=Last phase+Phase Increment}
	ax0 = ar;			{ The phase input to sin is reg ax0 }
	dm(phase) = ar;			{ Save for next data point }
	call sin;			{ Phase in ax0, Sin returned in ar }
	{ SIN_LEV is a constant multiplier of the amplitude of the sine wave
	  (i.e. an amplitude modulator) that sets the S/N.         }
        my0 = SIN_LEV;		{ Get multiplier, in 1.15 }
        mr = ar * my0 (ss);		{ New scaled sine wave }
        dm(signal_save) = mr1;		{ Here sine wave signal in sr1 }
        call getrnd;			{ 16-bit rand numb, GWN in af }
        ar = dm(signal_save);		{ Non-noise part }
        ar = ar + af;			{ Sig + Noise }

	dm(tx_buf+1) = ar;		{ Send S+N wave to Left D/A (Codec) }
	dm(tx_buf+2) = ar;		{ Right D/A }

	dis sec_reg;			{ Restore the primary registers }
	rti;				{ This undoes the interrupt }

{*************************************************************************
 *      getrnd    Random Number Generator - 8 uniform rn added together
 *                to approximate Gaussian rn.  RSL 14 Feb 96
 * Uniform routine from AD DSP Appl Vol 1.  The linear congruence method is
 * described in Kunth, D. E. "The Art of Computer Programming, Vol 2", Addison
 * Wesle, 1969.  This method multiplies the last random number by a constant
 * a, adds a second constant c and then takes the most significant 16 bits as
 * the new uniformly distributed random number that is uniform on the full
 * 16 bit range.  Taken as signed 1.15 format number this corresponds to -1.0
 * to +32767/32768 with zero mean.
 * After dividing by 8 to prevent numerical overflow, 8 of these uniform
 * random numbers are added together.  An important mathematical principal,
 * called the Central Limit theorem, tells us that as the number of independent
 * numbers added together gets large the result will closely approximate
 * a Gaussian distribution.  Eight gives a good approximation.
 * The sample output in af is in 1.15 format with sigma=0.20412 (sig^2= 1/24)
 * with mean=0;
 * Thus  peak/sigma is 1/.20412 = 4.9 which is sufficient for 99.99998% of
 * all Gaussian samples.
 *
 * Calling value:  dm(seed_msw), dm(seed_lsw)
 * Return value:   af is noise sample
 * Instructions:   134 instruction cycles per output
 *
 *****************************************************************************}
getrnd:	{ The constant a=1664525 decimal, and is represented here as two 16
	  bit values:       }
        my1=25;                         { Upper half of a (1664525/65536) }
        my0=26125;                      { Lower half of a, the remainder }
        af=pass 0;			{ Clear the arithmetic accumulator }
        cntr=8;				{ The number of uniform rn added }
	{ Now loop 8 times to generate a noise sample:   }
        do randloop until ce;		{ Decrease cntr until 0 }
	  sr1=dm(seed_msw);		{ Get the 32 bit seed from last }
	  sr0=dm(seed_lsw);		{   call to this fcn or last loop }

          mr=sr0*my1(uu);		{ 32 bit multiply: a(hi) * x(lo) }
          mr=mr+sr1*my0(uu);            {   and a(hi) * x(lo) + a(lo) X x(hi) }
          si=mr1;			{ Temp storage }
          mr1=mr0;
          mr2=si;
          mr0=h#fffe;                   { c=32767, left-shifted by 1 }
          mr=mr+sr0*my0(uu);            { (above) + a(lo) X x(lo) + c }
          sr=ashift mr2 by 15 (hi);
          sr=sr or lshift mr1 by -1 (hi); { Right-shift by 1 }
          sr=sr or lshift mr0 by -1 (lo); { Now have uniform rn in sr1 }
	  { Use seed_msw and seed_lsw as temporary storage.  This is as
	    efficient as borrowing some unused registers, and less risky when
	    program changes are made. Leaves next seed where we want it, also:}
	  dm(seed_msw)=sr1;		{ Save new seed value, high 16 bits }
	  dm(seed_lsw)=sr0;		{    and low 16 }
	  { Uniform random number still in sr1.  Add on to accumulator:  }
          sr=ashift sr1 by -3 (hi);	{ Divide by 8, i.e., shift right 3 }
randloop: af=sr1+af;			{ Accumulate 8 uniform rn }
	rts;				{ random 16-bit value in af }

{*************************************************************************
 *
 * Sine waves are calculated by a polynomial approximation.  The
 * calculation is done only for 0 to 90 degrees and then the symmetry
 * of the sine wave is used to find the other three quadrants.
 * The phase change from point to point is 1.365333 times the desired
 * frequency in Hz.  The phase change must be rounded to the nearest
 * integer.
 *
 *       Sine Approximation:     ar = Sin(ax0)
 *       Calling Parameters:     ax0 in scaled 1.15 format
 *       Return Values:          ar in 1.15 format
 *       Computation Time:       25 cycles
 * Thanks to Analog Devices Application Handbook #1
 *************************************************************************}
sin:    m4=1;    l4=0;
        i4=^sin_coeff;                  { Pointer to coeff. buffer }
        ay0=H#4000;
        ar=ax0, af=ax0 and ay0;         { Check 2nd or 4th quad. }
        if ne ar=-ax0;                  { If yes, negate input }
        ay0=H#7FFF;
        ar=ar and ay0;                  { Remove sign bit }
        my1=ar;
        mf=ar*my1 (RND), mx1=pm(i4,m4); { mf = input**2 }
        mr=mx1*my1 (SS), mx1=pm(i4,m4); { Start polynomial calculation }
        cntr=3;
        do approx until ce;
                mr=mr+mx1*mf (ss);
approx:         mf=ar*mf (rnd), mx1=pm(i4,m4);
        mr=mr+mx1*mf (ss);
        sr=ashift mr1 by 3 (hi);
        sr=sr or lshift mr0 by 3 (lo);  { Convert to 1.15 format }
        ar=pass sr1;
        if lt ar=pass ay0;              { Saturate if needed }
        af=pass ax0;
        if lt ar=-ar;                   { Negate output if needed }
        rts;

{*****************************************************************************
 * Main startup initialization.  Many variables and registers require
 * startup initialization, either for operation or to improve simulation.
 * This routine includes the parts that can easily be simulated
 ****************************************************************************}
init0:	ax0 = RN_SEED;
        dm(seed_lsw) = ax0;             { Arbitrary seed for rn generator }
        dm(seed_msw) = ax0;
	ax0=16384;
	dm(sin_lev)=ax0;

	{ i7, l7, i6, l6 devoted to auto-buffering in SPORT0.  Initialized
	  here. These need protection, if used elsewhere in the program.    }
        i7 = ^rx_buf;
        l7 = %rx_buf;
        m7 = 1;
        i6 = ^tx_buf;
        l6 = %tx_buf;
        m6 = 1;
	{ i5, l5, m5 used in next_cmd to load AD1847 regs and then made free. }
        i5 = ^init_cmds;
        l5 = %init_cmds;
        m5 = 1;

        ax0=0xC000;  dm(tx_buf)=ax0;    { Initially set MCE }
        ax0=0x0000;  dm(tx_buf+1)=ax0;
        ax0=0x0000;  dm(tx_buf+2)=ax0;


	{  AD1847  Command Register Initialization.  They are placed into
	   memory here and will be loaded into the AD1847 with interrupt
	   routines in init1.  The details of these registers are given
	   in the AD1847 data sheet.    }
        ax0=0xc002;  dm(init_cmds)=ax0;
        ax0=0xc102;  dm(init_cmds+1)=ax0;
        ax0=0xc288;  dm(init_cmds+2)=ax0;
        ax0=0xc388;  dm(init_cmds+3)=ax0;
        ax0=0xc488;  dm(init_cmds+4)=ax0;
        ax0=0xc588;  dm(init_cmds+5)=ax0;
        ax0=0xc680;  dm(init_cmds+6)=ax0;
        ax0=0xc780;  dm(init_cmds+7)=ax0;
        ax0=0xc85c;  dm(init_cmds+8)=ax0;  {48 KHz, c850h for 8 KHz }
        ax0=0xc909;  dm(init_cmds+9)=ax0;
        ax0=0xca00;  dm(init_cmds+10)=ax0;
        ax0=0xcc40;  dm(init_cmds+11)=ax0;
        ax0=0xcd00;  dm(init_cmds+12)=ax0;

{********************* Serial PORT 0 Initialization *************************}
        ax0= b#0000110101111111;   dm (SPORT0_Autobuf) = ax0;
        ax0 = 0;    dm (SPORT0_RFSDIV) = ax0;
        ax0 = 0;    dm (SPORT0_SCLKDIV) = ax0;
        ax0 = b#1000011000001111;   dm (SPORT0_Control_Reg) = ax0;
        ax0 = b#0000000000000111;   dm (SPORT0_TX_Channels0) = ax0;
        ax0 = b#0000000000000111;   dm (SPORT0_TX_Channels1) = ax0;
        ax0 = b#0000000000000111;   dm (SPORT0_RX_Channels0) = ax0;
        ax0 = b#0000000000000111;   dm (SPORT0_RX_Channels1) = ax0;


{******************* System and memory initialization ***********************}
	ax0 = 0;    dm(DM_Wait_Reg)=ax0;
        ax0 = b#0001000000000000;	{ Enable SPORT0 }
        dm(System_Control_Reg) = ax0;
        ifc = b#0000000011111111;	{ Clear any pending interrupt }
        nop;
	icntl = 0;
	mstat = b#1000000;

#if 0
ax0=0;
dm(overrun_test)=ax0;
ax0=7;
dm(overrun)=ax0;
#endif
	rts;

{*****************************************************************************
 * Start up initialization of the AD1847 Codec.  This requires some effort to
 * simulate, and it is easiest to skip this part when running the sim21
 * simulator.
 *****************************************************************************}
init1:

	{ AD 1847 CODEC Initialization
	  The stat_flag=0 will prevent reinit of AD1847 from SPORT0 TX ints }
	ax0 = 1;                        { set flag, disable tx ints }
	dm(stat_flag) = ax0;
	imask = b#0001000000;           { enable transmit interrupt }
	ax0 = dm(i6, m6);		{ Start interrupt }
	tx0 = ax0;
check_init:
        ax0 = dm (stat_flag);		{ wait for entire init }
        af = pass ax0;			{ buffer to be sent to }
        if ne jump check_init;		{ the codec            }
        ay0 = 2;
check_acih:
        ax0 = dm (rx_buf);          { once initialized, wait for codec }
        ar = ax0 and ay0;           { to come out of autocalibration }
        if eq jump check_acih;      { wait for bit set }
check_aci1:
        ax0 = dm (rx_buf);          { wait for bit clear }
        ar = ax0 and ay0;
        if ne jump check_aci1;
	idle;
        ay0 = 0xbf3f;               { unmute left DAC }
        ax0 = dm (init_cmds + 6);
        ar = ax0 AND ay0;
        dm (tx_buf) = ar;
        idle;
        ax0 = dm (init_cmds + 7);   { unmute right DAC }
        ar = ax0 AND ay0;
        dm (tx_buf) = ar;
        idle;
	ax0 = 0xc901;			{ Clear autocalibration request }
	dm(tx_buf)=ax0;
	idle;
        ax0 = 0x8000;               { control word to clear over-range flags }
        dm (tx_buf) = ax0;
        idle;

        ifc = b#00000011111111;     { clear any pending interrupt }
        nop;
        imask = b#0000100000;       { enable rx0 interrupt }
	rts;


{************************  NEXT_CMD  ****************************************
 * Transmit Interrupt used for Codec initialization
 * DAG 5 registers used here, but are available after this routine is run.
 * Adapted from EZ-KIT Lite routines.
 ****************************************************************************}
next_cmd:
        ena sec_reg;
        ax0 = dm (i5, m5);          { fetch next control word and }
        dm (tx_buf) = ax0;          { place in transmit slot 0    }
        ax0 = i5;
        ay0 = ^init_cmds;
        ar = ax0 - ay0;
        if gt rti;                  { rti if more control words still waiting }
        ax0 = 0x8000;               { else set done flag and }
        dm (tx_buf) = ax0;          { remove MCE if done initialization }
        ax0 = 0;
        dm (stat_flag) = ax0;       { reset status flag }
        dis sec_reg;
        rti;

.endmod;
